#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
The utils module contains utility functions.
These include tools from primary utilities (get_pipes)
to miscellaneous helper functions.
"""

### import_children will be depreciated in a future release.
### Explicit is better than implicit (mostly).

# from meerschaum.utils.packages import import_children
# import_children()

from meerschaum.utils.get_pipes import get_pipes
